import math
import torch
import torch.nn as nn
from functools import partial
from timm.models.vision_transformer import trunc_normal_
from .masked_autoencoder import MaskedAutoencoder
from .models_hivit import HiViT, PatchEmbed, PatchMerge, BlockWithRPE
from util.pos_embed import get_2d_sincos_pos_embed


class HiViTMaskedAutoencoder(MaskedAutoencoder, HiViT):
    def __init__(self, img_size=224, patch_size=16, in_chans=3,
                 embed_dim=512, depths=[2, 2, 20], num_heads=8, stem_mlp_ratio=3., mlp_ratio=4., 
                 qkv_bias=True, qk_scale=None, drop_rate=0., attn_drop_rate=0., drop_path_rate=0.0,
                 norm_layer=nn.LayerNorm, ape=True, rpe=True, patch_norm=True, use_checkpoint=False, 
                 decoder_embed_dim=512, decoder_depth=8, decoder_num_heads=16, hifeat=False,
                 **kwargs):
        MaskedAutoencoder.__init__(self)
        self.num_layers = len(depths)
        self.ape = ape
        self.rpe = rpe
        self.patch_norm = patch_norm
        self.num_features = embed_dim
        self.mlp_ratio = mlp_ratio
        self.use_checkpoint = use_checkpoint
        self.num_main_blocks = depths[-1]
        self.hifeat = hifeat

        embed_dim = embed_dim // 2 ** (self.num_layers - 1)
        # split image into non-overlapping patches
        self.patch_embed = PatchEmbed(
            img_size=img_size, patch_size=patch_size, in_chans=in_chans, embed_dim=embed_dim,
            norm_layer=norm_layer if self.patch_norm else None)
        num_patches = self.patch_embed.num_patches
        Hp, Wp = self.patch_embed.patches_resolution
        assert Hp == Wp

        # absolute position embedding
        if ape:
            self.absolute_pos_embed = nn.Parameter(
                torch.zeros(1, num_patches, self.num_features, requires_grad=False)
            )
            trunc_normal_(self.absolute_pos_embed, std=.02)
        if rpe:
            # get pair-wise relative position index for each token inside the window
            coords_h = torch.arange(Hp)
            coords_w = torch.arange(Wp)
            coords = torch.stack(torch.meshgrid([coords_h, coords_w])) 
            coords_flatten = torch.flatten(coords, 1) 
            relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :] 
            relative_coords = relative_coords.permute(1, 2, 0).contiguous() 
            relative_coords[:, :, 0] += Hp - 1 
            relative_coords[:, :, 1] += Wp - 1
            relative_coords[:, :, 0] *= 2 * Wp - 1
            relative_position_index = relative_coords.sum(-1)
            self.register_buffer("relative_position_index", relative_position_index)

        self.pos_drop = nn.Dropout(p=drop_rate)

        # stochastic depth
        dpr = iter(x.item() for x in torch.linspace(0, drop_path_rate, sum(depths) + sum(depths[:-1])))

        # build blocks
        self.blocks = nn.ModuleList()
        for stage_depth in depths:
            is_main_stage = embed_dim == self.num_features
            nhead = num_heads if is_main_stage else 0
            ratio = mlp_ratio if is_main_stage else stem_mlp_ratio
            # every block not in main stage include two mlp blocks
            stage_depth = stage_depth if is_main_stage else stage_depth * 2
            for _ in range(stage_depth):
                self.blocks.append(
                    BlockWithRPE(
                        Hp, embed_dim, nhead, ratio, qkv_bias, qk_scale, 
                        drop=drop_rate, attn_drop=attn_drop_rate, drop_path=next(dpr), 
                        rpe=rpe, norm_layer=norm_layer,
                    )
                )
            if not is_main_stage:
                self.blocks.append(
                    PatchMerge(embed_dim, norm_layer)
                )
                embed_dim *= 2

        self.num_features = 7 * embed_dim if self.hifeat else embed_dim
        self.norm = norm_layer(self.num_features)
        # --------------------------------------------------------------------------

        # --------------------------------------------------------------------------
        # MAE decoder specifics
        self.decoder_patch_size = patch_size
        self.decoder_embed = nn.Linear(self.num_features, decoder_embed_dim, bias=True)

        self.mask_token = nn.Parameter(torch.zeros(1, 1, decoder_embed_dim))

        self.decoder_pos_embed = nn.Parameter(torch.zeros(1, num_patches, decoder_embed_dim), requires_grad=False)  # fixed sin-cos embedding

        self.decoder_blocks = nn.ModuleList([
            BlockWithRPE(
                Hp, decoder_embed_dim, decoder_num_heads, mlp_ratio, qkv_bias, qk_scale, 
                rpe=False, norm_layer=norm_layer
            )
            for _ in range(decoder_depth)])

        self.decoder_norm = norm_layer(decoder_embed_dim)
        self.decoder_pred = nn.Linear(decoder_embed_dim, self.decoder_patch_size**2 * in_chans, bias=True) # decoder to patch
        # --------------------------------------------------------------------------

        # initialize (and freeze) pos_embed by sin-cos embedding
        pos_embed = get_2d_sincos_pos_embed(self.absolute_pos_embed.shape[-1], Hp, cls_token=False)
        self.absolute_pos_embed.data.copy_(torch.from_numpy(pos_embed).float().unsqueeze(0))

        decoder_pos_embed = get_2d_sincos_pos_embed(self.decoder_pos_embed.shape[-1], Hp, cls_token=False)
        self.decoder_pos_embed.data.copy_(torch.from_numpy(decoder_pos_embed).float().unsqueeze(0))
        torch.nn.init.normal_(self.mask_token, std=.02)

        self.apply(self._init_weights)
    
    def masking_id(self, batch_size, mask_ratio):
        N, L = batch_size, self.absolute_pos_embed.size(1)
        len_keep = int(L * (1 - mask_ratio))
        
        noise = torch.rand(N, L, device=self.absolute_pos_embed.device)  # noise in [0, 1]

        # sort noise for each sample
        ids_shuffle = torch.argsort(noise, dim=1)  # ascend: small is keep, large is remove
        ids_restore = torch.argsort(ids_shuffle, dim=1)

        # keep the first subset
        ids_keep = ids_shuffle[:, :len_keep]
        # generate the binary mask: 0 is keep, 1 is remove
        mask = torch.ones([N, L], device=self.absolute_pos_embed.device)
        mask[:, :ids_keep.size(1)] = 0
        # unshuffle to get the binary mask
        mask = torch.gather(mask, dim=1, index=ids_restore)

        return ids_keep, ids_restore, mask
    
    def forward_encoder(self, x, mask_ratio):
        ids_keep, ids_restore, mask = self.masking_id(x.size(0), mask_ratio)

        if self.hifeat:
            x = self.forward_features(x, ids_keep=ids_keep, return_hifeat=True)
            h, m, l = x
            B, N, _ = l.shape
            x = torch.cat([h.reshape(B, N, -1), m.reshape(B, N, -1), l], dim=-1)
            x = self.norm(x)
        else:
            x = self.forward_features(x, ids_keep=ids_keep)

        return x, mask, ids_restore
    
    def forward_decoder(self, x, ids_restore):
        # embed tokens
        x = self.decoder_embed(x)

        # append mask tokens to sequence
        mask_tokens = self.mask_token.repeat(x.shape[0], ids_restore.shape[1] + 1 - x.shape[1], 1)
        x = torch.cat([x, mask_tokens], dim=1)  
        x = torch.gather(x, dim=1, index=ids_restore.unsqueeze(-1).repeat(1, 1, x.shape[2]))

        # add pos embed
        x = x + self.decoder_pos_embed

        # apply Transformer blocks
        for blk in self.decoder_blocks:
            x = blk(x)
        x = self.decoder_norm(x)

        # predictor projection
        x = self.decoder_pred(x)
        return None, x


def mae_hivit_tiny_dec512d8b(**kwargs):
    model = HiViTMaskedAutoencoder(
        embed_dim=384, depths=[1, 1, 10], num_heads=6, stem_mlp_ratio=3., mlp_ratio=4., 
        decoder_embed_dim=512, decoder_depth=8, decoder_num_heads=16,
        rpe=False, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model


def mae_hivit_small_dec512d8b(**kwargs):
    model = HiViTMaskedAutoencoder(
        embed_dim=384, depths=[1, 1, 22], num_heads=6, stem_mlp_ratio=3., mlp_ratio=4., 
        decoder_embed_dim=512, decoder_depth=8, decoder_num_heads=16,
        rpe=False, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model


def mae_hivit_base_dec512d8b(**kwargs):
    model = HiViTMaskedAutoencoder(
        embed_dim=512, depths=[2, 2, 20], num_heads=8, stem_mlp_ratio=3., mlp_ratio=4., 
        decoder_embed_dim=512, decoder_depth=8, decoder_num_heads=16,
        rpe=False, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model


def mae_hivit_base_dec512d8b_hifeat(**kwargs):
    model = HiViTMaskedAutoencoder(
        embed_dim=512, depths=[2, 2, 20], num_heads=8, stem_mlp_ratio=3., mlp_ratio=4., 
        decoder_embed_dim=512, decoder_depth=8, decoder_num_heads=16, hifeat=True,
        rpe=False, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model


def mae_hivit_base_dec512d8b_hifeat_rpe(**kwargs):
    model = HiViTMaskedAutoencoder(
        embed_dim=512, depths=[2, 2, 20], num_heads=8, stem_mlp_ratio=3., mlp_ratio=4., 
        decoder_embed_dim=512, decoder_depth=8, decoder_num_heads=16, hifeat=True,
        rpe=True, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model


def mae_hivit_base_dec512d6b(**kwargs):
    model = HiViTMaskedAutoencoder(
        embed_dim=512, depths=[2, 2, 20], num_heads=8, stem_mlp_ratio=3., mlp_ratio=4., 
        decoder_embed_dim=512, decoder_depth=6, decoder_num_heads=16,
        rpe=False, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model


def mae_hivit_base_dec512d4b(**kwargs):
    model = HiViTMaskedAutoencoder(
        embed_dim=512, depths=[2, 2, 20], num_heads=8, stem_mlp_ratio=3., mlp_ratio=4., 
        decoder_embed_dim=512, decoder_depth=4, decoder_num_heads=16,
        rpe=False, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model


def mae_hivit_base_dec512d2b(**kwargs):
    model = HiViTMaskedAutoencoder(
        embed_dim=512, depths=[2, 2, 20], num_heads=8, stem_mlp_ratio=3., mlp_ratio=4., 
        decoder_embed_dim=512, decoder_depth=2, decoder_num_heads=16,
        rpe=False, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model


def mae_hivit_base_dec512d1b(**kwargs):
    model = HiViTMaskedAutoencoder(
        embed_dim=512, depths=[2, 2, 20], num_heads=8, stem_mlp_ratio=3., mlp_ratio=4., 
        decoder_embed_dim=512, decoder_depth=1, decoder_num_heads=16,
        rpe=False, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model


def mae_hivit_large_dec512d8b(**kwargs):
    model = HiViTMaskedAutoencoder(
        embed_dim=768, depths=[2, 2, 40], num_heads=12, stem_mlp_ratio=3., mlp_ratio=4., 
        decoder_embed_dim=512, decoder_depth=8, decoder_num_heads=16,
        rpe=False, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model


def mae_hivit_large_dec512d8b_hifeat(**kwargs):
    model = HiViTMaskedAutoencoder(
        embed_dim=768, depths=[2, 2, 40], num_heads=12, stem_mlp_ratio=3., mlp_ratio=4., 
        decoder_embed_dim=512, decoder_depth=8, decoder_num_heads=16, hifeat=True,
        rpe=False, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model


def mae_hivit_large2_dec512d8b_hifeat(**kwargs):
    model = HiViTMaskedAutoencoder(
        embed_dim=768, depths=[4, 4, 36], num_heads=12, stem_mlp_ratio=3., mlp_ratio=4., 
        decoder_embed_dim=512, decoder_depth=8, decoder_num_heads=16, hifeat=True,
        rpe=False, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model


def mae_hivit_large_dec512d8b_hifeat_rpe(**kwargs):
    model = HiViTMaskedAutoencoder(
        embed_dim=768, depths=[2, 2, 40], num_heads=12, stem_mlp_ratio=3., mlp_ratio=4., 
        decoder_embed_dim=512, decoder_depth=8, decoder_num_heads=16, hifeat=True,
        rpe=True, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model


def mae_hivit_large1024_dec512d8b(**kwargs):
    model = HiViTMaskedAutoencoder(
        embed_dim=1024, depths=[2, 2, 20], num_heads=16, stem_mlp_ratio=3., mlp_ratio=4., 
        decoder_embed_dim=512, decoder_depth=8, decoder_num_heads=16,
        rpe=False, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model


mae_hivit_base = mae_hivit_base_dec512d8b_hifeat
mae_hivit_large = mae_hivit_large2_dec512d8b_hifeat
